//!The RAq Approach: cluster.h
/*!
Contains many of the intermediate functions necessary for the decomposition
and merging strategies needed for building the decomposition tree and the final
supertree.
*/
//(c) 2006 Suzanne Matthews
/*
This file is part of the RaqApproach software package.

    RaqApproach is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    RaqApproach is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with RaqApproach; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef CLUSTER_H_
#define CLUSTER_H_

#include "elements.h"
#include "tree.h"
#include "common.h"

using namespace std;

int create_tree(multimap<double, string>my_matrix); //!< starter function for decomposition step: creates decomposition tree
void create_supertree(list<Node *> & my_leaves ); //!< starter function for merger step: creates supertree

void build_tree(Tree &newtree, list<element> cluster, double q) { //!< build_tree recursively creates tree (decomposition step)
/*using a lambda value (of 2), the buid_tree() function recursively breaks down the current
  element set and add it to our decomposition
  tree*/
	double q_ = 0; //!< counter for current q value. starts at 0
	double new_q = q/(float)decay; //!< value that q_ is incremented by per iteration
	bool equidistant = false; //!<tells us when a set of clustered elements cannot be broken down any further
	Node * temp_current = newtree.pointer_current();
	list<element> temp;
	list<element>::iterator e_itr = cluster.begin();
	bool first_pass = 1; //!< tells us when we should perform a "left" insert, as opposed to a "right" insert


	double check_dist = (*e_itr).get_dist();
	while ( e_itr != cluster.end() ) { //!< code that will check to see if the distances are the same

		if ( (*e_itr).get_dist() != check_dist)
			break;
		++e_itr;
	}
	if ( e_itr == cluster.end() ) //!< if we reach the end of the vector
		equidistant = true; //!< set equidistant to true
	e_itr = cluster.begin();

	while ( e_itr != cluster.end() ) { //!< while there are still elements remaining in the cluster

		if ( (*e_itr).get_dist() > q_ ) { //!< if the current distance is greater than the current threshold

             if ( equidistant ) { //!< if the elements in our cluster are equidistant
				 list<element>::iterator temp_itr =  cluster.begin();
				 while ( temp_itr != cluster.end() ) {
					 temp.push_back((*e_itr)); //!<add them to temp. must add to tree, since can't be broken down further
					 ++temp_itr;
				 }
			 }
			if ( !temp.empty() ) {
				if ( first_pass ) { //!<if this is the first pass
					newtree.insert(temp, 0); //!<insert on left
					newtree.left();
					first_pass = 0; //!<set first_pass to false
				}
				else {
					newtree.insert(temp,1); //!<otherwise, insert on right
					newtree.right();
				}
				if ( (temp.size() > lambda) && (!equidistant) ) { //!<if it is not equidistant and is above the threshold
					Node * new_current = newtree.pointer_current();
					build_tree(newtree, temp, new_q); //!<call build tree again on the cluster
					newtree.SetCurrent(new_current);
				}

 				temp.clear(); //!<clear the temp vector

			}
			q_ += new_q; //!<increment the threshold by the requisite amount

		} //!< end greater than q_
		else {
			temp.push_back((*e_itr)); //!<else, add to the temporary cluster vector
			++e_itr;
		}

	} //!<end while loop
	if ( !temp.empty() ) { //!< if there are  unclustered elements
		 if ( first_pass ) { //!<if it is the first pass
			 newtree.insert(temp,0); //!<insert on left
			 newtree.left();
		 }
		 else {
			newtree.insert(temp, 1); //!<else insert on right
			newtree.right();
		}
		if ( (temp.size() > lambda) && (!equidistant) ) { //!<if the unclustered elements are not equidistant and above threshold
			Node * last = newtree.pointer_current();
			build_tree(newtree, temp, new_q); //!< call build tree again on the cluster
			newtree.SetCurrent(last);
		}
		temp.clear(); //!< clear the temporary cluster array.
	}
	nid temp_id = temp_current->id;
	nid my_id = newtree.pointer_current()->id;
	if ( DISPLAY ) {
		cout << "my id is: " << endl;
		for ( unsigned int i = 0; i < my_id.size(); ++i)
				cout << " | " << my_id[i];
		cout << endl;
		cout << "set parent to node with the following id: " << endl;
		for ( unsigned int i = 0; i < temp_id.size(); ++i)
			cout << " | " << temp_id[i];
		cout << endl;
	}
	newtree.SetCurrent(temp_current); //set the current node back up to the node where the recursion began
} //end function

int create_tree(multimap<double, string> my_matrix) {

  vector<int> recognition(num_taxa, 0); //!<recognition array
  list<element> clustered; //!<this will be the input for the decomposition function
  /*create_tree accepts an inputted matrix and from it creates a tree structure in newick format*/
  ofstream fout;
  fout.open("graph_me.txt"); //!<this file is for cluster benchmarking (preliminary)
  double q_heat = initial_q; //!<this is our dynamic q value
  fout << "for heating value of: " << q_heat << endl;
  fout << "cluster size" << endl;
  float q = 0; //!<it will be repeatedly "heated" to yield subtrees in a certain range

  multimap<double, string>::iterator m_itr = my_matrix.begin();
  int count_clusters = 1;
  int largest = 0;
  int smallest = 10000; //!<default value
  bool first_pass = 1;
  Tree my_tree;  //!< create a tree
  my_tree.insert(clustered, 0);  //!<  dummy vector placed at root
  clustered.clear();
  while (m_itr != my_matrix.end()) {  //!< while we still have elements in the matrix to process

    if (m_itr->first > q ){ //!< if the current distance is greater than the current threshold
      if (!clustered.empty()) { //!< if the cluster is not empty
         if (first_pass) { //!< if it is the first pass
			  my_tree.insert(clustered, 0); //!< insert on left
			  my_tree.left(); //!<  have current pointer point to that element
			  first_pass = 0; //!<  set first_pass to false
		  }
		  else {
			  my_tree.insert(clustered, 1); //!<  else, insert on right
			  my_tree.right(); //!< have current pointer point to that element
			}

		  if ( clustered.size() > lambda ) { //!<  if the cluster size is greater than our horizontal threshold
			  Node * reset_node = my_tree.pointer_current();
			  build_tree(my_tree, clustered, q_heat); //!<  call build_tree() function, which will recursively break down clusters
			  my_tree.SetCurrent(reset_node);
		  }

          fout << count_clusters << " " << clustered.size() << endl;
		  if ( DISPLAY ) {
			  cout << "cluster " << count_clusters << ": " << endl;
			  list<element>::iterator itr = clustered.begin();
			  while ( itr != clustered.end() ) {
				  cout << (*itr).get_taxa() << endl;
				  ++itr;
			  }
		  }

	  	  if (clustered.size() > largest) { //!<  note the largest cluster found so far
			  largest = clustered.size();
		  }
	 	  if (clustered.size() < smallest) { //!<  not the smallest cluster found so far
			  smallest = clustered.size();
		  }

		  ++count_clusters; //!<  count the number of preliminary clusters (for top level)
		  clustered.clear(); //!<  clear cluster array
	 }

     q+=q_heat; //!< increment threshold value by "heat" value

    }
    else { //!<  else, if the current distance is below the threshold
		string temp = m_itr->second;
		string temp2;
		string::iterator t_itr= temp.begin();
		bool done = false;
		//!< parse the first element
		while ( t_itr != temp.end() ) {

			if ( (*t_itr) == '.' ) {
				temp.erase(t_itr);
				done = true;
				break;
			}
			else {
				temp2.push_back((*t_itr));
				temp.erase(t_itr);
			}
			if ( done == true )
				break;

		}

		int first = 0, second = 0;
		stringstream ss1, ss2;
		ss1 << temp;
		ss1 >> first;
		ss2 << temp2;
		ss2 >> second;
        /*! the recognition vector is vital to ensuring that no more than two appearances
          of each taxa ever show up in our final clusters*/
		temp.clear(); temp2.clear();
		int tally = 0; //!<  tally keeps track of which taxa have been added to the tree already
		if ( recognition[first-1] != 0 ) {
			tally += recognition[first-1];
			//if ( recognition[first-1] > 2 )
			//	recognition[first-1] == 2;
		}
		//else
			//recognition[first-1] = recognition[first-1] + 1;

		if ( recognition[second-1] != 0 ) {
			tally += recognition[second-1];
			//if ( recognition[second-1] > 2 )
			//	recognition[second-1] == 2;
		}
		//else
			//recognition[second-1] = recognition[second-1] + 1;

	  	if (tally < 2 ) { //!<  check to see if it's ok to cluster the element.. must be less than 2 (nothing to do with lambda)
			element temp(m_itr->second, m_itr->first);
	  		clustered.push_back(temp);
	  		recognition[first-1] += 1;
	  		recognition[second-1]+= 1;
		}
	    ++m_itr; //!< go to next element in matrix
    }
  } //!< end of while loop

  if ( !clustered.empty() ) { //!< if there are any unclustered elements
	  if ( first_pass ) { //!< if it is the first pass
		  my_tree.insert(clustered, 0); //!<  insert on left
		  my_tree.left();
	  }
	  else { //!<  else, insert on right
		my_tree.insert(clustered, 1);
		my_tree.right();
	  }
	  if ( clustered.size() > lambda ) { //!<  if this cluster is greater than our threshold
		Node * last = my_tree.pointer_current();
		build_tree(my_tree, clustered, q_heat); //!<  call the build_tree() function
		my_tree.SetCurrent(last);
	  }
	  clustered.clear(); //!<  clear the cluster vector
  }
  //!<  print out statistics for cluster size (optional)
  cout << "Some statistics (for initial cluster size)..." << endl;
  cout << "number of clusters: " << count_clusters << endl;
  cout << "largest cluster size is: " << largest << endl;
  cout << "smallest cluster size is: " << smallest << endl;
  //!<  print out uncluster elements (a check, also optional)
  cout << "unclustered elements: " << endl;
  for (unsigned int i = 0; i < recognition.size(); ++i) {
	  if ( recognition[i] < 1 )
	  	cout << i+1 << endl;
	}
  cout << endl;
  my_tree.reset();
  if ( DISPLAY ) {
  	show_leaves(my_tree.pointer_current()); //!<  display the leaves
  	my_tree.reset();
  }
  list<Node *> my_leaves;
  find_leaves(my_leaves, my_tree.pointer_current()); //!< add the leaves to the my_leaves vector
  if ( DISPLAY ) {
	  cout << endl << endl << endl << "LEAF NODES: " << endl;
	  list<Node *>::iterator l_itr = my_leaves.begin(); //!< print out the leaf nodes
	  while ( l_itr != my_leaves.end() ) {
		  display_node(*l_itr);
		  ++l_itr;
	  }
  	  cout << endl << endl << endl << "ALL NODES: " << endl;
  	  //! print out all the nodes in the decomposition tree
  	  my_tree.DisplayPreorder(my_tree.pointer_current());
  }
  create_supertree(my_leaves); //!<  call the super_tree() method
  cout << "super tree creation complete." << endl;
  return 0;
}


vector<int> separate(string taxon) { //!<  separates a string of the form A.B into the two taxa A and B
	vector<int> my_taxa;
	int i = 0;
	string s1, s2;
	while (taxon[i] != '.') {
		s1.push_back(taxon[i]);
		++i;
	}
	if ( taxon[i] == '.' )
		++i;
	for ( i; i < taxon.size(); ++i ) {
		s2.push_back(taxon[i]);
	}

	int first = 0, second = 0;
	stringstream ss1, ss2;
	ss1 << s1;
	ss1 >> first;
	ss2 << s2;
    ss2 >> second;

	my_taxa.push_back(first);
	my_taxa.push_back(second);

	return my_taxa; //!< returns a vector containing the two now separated taxa
}

map<int, list<int> > create_hash(list<Node *> my_leaves) { //!< creates the hash table necessary for intial merge step (cluster merge)

	map<int, list<int> > hash_map; //!<  declare hash map
	list<Node *>::iterator sift = my_leaves.begin();
	while ( sift != my_leaves.end() ) { //!<  while there are still leaves to process
		list<element> clust = (*sift)->data; //!<  grab the first leaf
		list<element>::iterator c_itr = clust.begin();
		while ( c_itr != clust.end() ) { //!<  while there are still elements remaining in the leaf
			string temp = (*c_itr).get_taxa();
			vector<int> taxas = separate(temp); //!<  separate the taxa
			for ( unsigned int i = 0; i < taxas.size(); ++i) {
				map<int, list<int> >::iterator hasher;
				hasher = hash_map.find(taxas[i]); //!<  look for taxa in hash table
				if ( hasher != hash_map.end() ) { //!<  if it's in there
					if ( i == 0 ) { //!<  depending on the iteration
						hash_map[taxas[i]].push_back(taxas[1]); //!<  add it the corresponding place
					}
					else {
						hash_map[taxas[i]].push_back(taxas[0]); //!<  add it to the corresponding place
					}
				}
				else { //!<  if not, add the element to the hash table
					list<int> new_list;
					if ( i == 0 ) {
						new_list.push_back(taxas[i+1]);
					}
					else
						new_list.push_back(taxas[i-1]);
					hash_map.insert( make_pair(taxas[i],new_list) );
				}
			}
			++c_itr; //!<  go to next element in leaf
		}
		++sift; //!<  go to next leaf
	}

	return hash_map; //!<  return completed hash table
}

map<vector<short>, list<SuperTree> > create_superhash(list<SuperTree> subtrees, int &  longest) {
	//!<  hash table for final tree construction
	map<vector<short>, list<SuperTree> > hash_map;
	map<vector<short>, list<SuperTree> >::iterator hasher;
	list<SuperTree>::iterator place = subtrees.begin();
	longest = 0; //!<  keeps track of longest id found so far
	while ( place != subtrees.end() ) {
		(*place).reset(); //!< go to root
		vector<short> id = (*place).Id(); //!< grab the root node's ID
		if (id.size() > longest)
			longest = id.size(); //!<  find longest id thus far
		hasher = hash_map.find(id); //!<  search for the id
		if ( hasher != hash_map.end() ) { //!<  if we found the corresponding id
			hash_map[id].push_back((*place)); //!< add that tree to the list
		}
		else { //!<  else create a new entry for the hash table
			list<SuperTree> new_list;
			new_list.push_back((*place));
			hash_map.insert(make_pair(id, new_list)); //!<  insert into hash table
		}
		++place;
	}

	return hash_map; //!<  return hash table
}

void merge(int starter, map<int, list<int> > & my_map, vector<int> & product) { //!<  merge function for supertree subtree creation
	bool push = true;
	for ( unsigned int i = 0; i < product.size(); ++i) { //!<  determine whether or not we should add an element
		if ( product[i] == starter)
		push = false;
	}
	if (push)
		product.push_back(starter); //!<  if so, push_back()

	if ( my_map[starter].empty() ) //!<  if empty, return
		return;
	else {
		while ( !my_map[starter].empty() ) { //!<  while not empty
			int new_starter = my_map[starter].front(); //!<  get the first element
			my_map[starter].pop_front(); //!<  pop it from the front of the list
			merge(new_starter, my_map, product); //!<  call merge function again
		}
	}
}

void create_subtrees(vector<helper> help, list<SuperTree> & all_trees) { //!<  create subtrees function
	for ( unsigned int i = 0; i < help.size(); ++i ) {
		vector<int> temp_tree = help[i].get_tree(); //!<  temporary tree variable
		vector<short> temp_id = help[i].get_id(); //!<  temporary id variable
		int internal = temp_tree.size() - 1; //!<  number of internal nodes
		SuperTree new_tree;//!<  create supertree object
		new_tree.insert(0, 0, temp_id, true); //!<  create supernode object: internal node
		int taxa_ = temp_tree.back();
		temp_tree.pop_back();
		new_tree.insert(taxa_, 1, temp_id, false); //!<  create actual node. insert intial supernode into tree
		internal--; //!<  decrement value of internal
		while ( internal > 0 ) { //!<  while there are internal nodes remaining
			new_tree.insert(-1, 0, temp_id, true); //!<  add new internal node to left
			new_tree.left();
			taxa_ = temp_tree.back(); //!<  grab a taxa from the list
			temp_tree.pop_back(); //!<  remove the taxa
			new_tree.insert(taxa_, 1, temp_id, false); //!<  add that supernode to the right
			internal--; //!<  decrement number of internal nodes
		}
		taxa_ = temp_tree.back();
		temp_tree.pop_back();
		new_tree.insert(taxa_, 0, temp_id, false); //!<  insert remaining taxa on the left
		new_tree.left();
		//cout << endl;
		new_tree.reset();
		all_trees.push_back(new_tree); //!<  add the new tree to the list of subtrees
		new_tree.clear(); //!<  clear the new_tree from memory
	}
	return;
}

//!  displays a hash table
void display_hash( map<vector<short>, list<SuperTree> > merge_hash ) {
	map<vector<short>, list<SuperTree> >::iterator show = merge_hash.begin();
		while ( show != merge_hash.end() ) {
			vector<short> temp_id = show->first;
			for ( int i = 0; i < temp_id.size(); ++i) {
				cout << temp_id[i] << "|";
			}
			cout << " :||: ";
			list<SuperTree> temp = show->second;
			list<SuperTree>::iterator stuff = temp.begin();
			while ( stuff != temp.end() ) {
				(*stuff).DisplayInorder( (*stuff).root() );
				++stuff;
				cout << "**********************************************" << endl;
			}
			cout << endl;

			++show;
	}
}

 //!  merges subtrees into final tree (final step!)
void merge_supertrees(map<vector<short>, list<SuperTree> > hash, int longest ) {
	//!  merge_supertrees is a two step process:

	//!  during the first step, we iteratively go through the list and merge any list with two or more elements
	while ( longest > 1 && hash.size() != 1 ) {
		map<vector<short>, list<SuperTree> >::iterator merger = hash.begin();
		map<vector<short>, list<SuperTree> >::iterator finder;
		while ( merger != hash.end() ) {
			if ( merger->second.size() > 1 ) { //!<  if there are two or more trees hashed to any id
				list<SuperTree> temp = merger->second;
				merge_trees(temp); //!< merge those trees together
				merger->second.clear();
				merger->second = temp;
				temp.clear();
			}
			++merger;
		}

		if ( DISPLAY ) //!<  we display the resulting hash table (this is optional)
			display_hash(hash);

		//!< then we go through again, and remove the end of the ids that are equal to longest, and rehash them
		merger = hash.begin();
		if ( DISPLAY )
			cout << "STAGE TWO OF MERGE!!!!" << endl;
		while ( merger != hash.end() ) {
			if ( merger->first.size() == longest) { //!<  if we find an id that is equal to the longest
				vector<short> temp_id = merger->first;
				list<SuperTree> temp_list = merger->second;
				map<vector<short>, list<SuperTree> >::iterator temp_itr = merger;
				++merger;
				hash.erase(temp_itr);
				//!<  modify IDs
				temp_id.pop_back(); //!<  this is for the hash_table
				temp_list.front().root()->id.pop_back(); //!<  edit the id of tree in the list associated with id
				finder = hash.find(temp_id); //!<  look for the new id
				if ( finder != hash.end() ) {
					hash[temp_id].push_back(temp_list.front()); //!<  if new id is there, then add it to the corresponding list
				}
				else {
					hash.insert(make_pair(temp_id, temp_list)); //!<  else create a new element in hash table, and insert
				}
			}
			else
				++merger; //!<  increment merger
		}
		--longest; //!<  decrement the value of longest

		if ( DISPLAY ) {
			display_hash(hash);
			cout << "longest is: " << longest << endl;
			cout << "hash table's size is: " << hash.size() << endl;
		}
	}

    map<vector<short>, list<SuperTree> >::iterator merger = hash.begin();
    while ( merger != hash.end() ) { //!<  final step:
		if ( merger->second.size() > 1 ) { //!<  if there are multiple trees in our one element in the hash table
			list<SuperTree> temp = merger->second;
			merge_trees(temp); //!<  merge those trees together
			merger->second.clear();
			merger->second = temp;
			temp.clear();
		}
		++merger; //!<  do with all trees
	}

	merger = hash.begin();
	SuperTree checker = merger->second.front();
	list<string> print;
	checker.Newick(print, checker.root()); //!<  print out tree in Newick format
	print.push_back(";");
	cout << "The outputted tree is:" << endl;
	ofstream tout;
	tout.open("newick.txt");
	int count_taxa = 0;
	list<string>::iterator show = print.begin();
	while (show != print.end() ) {
		cout << (*show) << " ";
		tout << (*show); //!<  print Newick output to file "newick.txt"
		if ( ((*show)!=")") && ((*show)!="(") && ((*show)!=",") ) {
			++count_taxa;
		}
		++show;
	}
	cout << endl;
	tout.close();
	cout << "we have successfully created a tree with " << num_taxa << " taxa." << endl;

}

void create_supertree(list<Node *> & my_leaves ) { //!<  starting function for super_tree creation

	list<SuperTree> subtrees;

	//!<  declare hashmap
	map<int, list<int> > hash_map;
	hash_map = create_hash(my_leaves); //!<  create hash table

	//debugging code: print out hash_map:
	//cout << "Printing out hash_map!" << endl;
	if  ( DISPLAY ) {
		map<int, list<int> >::iterator show = hash_map.begin();
		while ( show != hash_map.end() ) {
			cout << show->first << " | ";
			list<int> temp = show->second;
			list<int>::iterator stuff = temp.begin();
			while ( stuff != temp.end() ) {
				cout << (*stuff) << " ";
				++stuff;
			}
			cout << endl;
			++show;
		}
	}

	list<Node *>::iterator rake = my_leaves.begin();
	vector< helper > temp_clusters; //!<  we will store the merged stuff in here
	while ( rake != my_leaves.end() ) { //!<  while there are still leaves remaining
		list<element> group = (*rake)->data; //!<  grab a component from one of the leaves
		vector<short> id = (*rake)->id;
		list<element>::iterator sorter = group.begin();
		while ( sorter != group.end() ) {
			string taxa = (*sorter).get_taxa(); //!<  grab one of the taxa
			vector<int> taxon = separate(taxa); //!<  separate the taxa
			vector<int> merged;
			if (!hash_map[taxon[0]].empty()) {
				merge(taxon[0], hash_map, merged); //!<  do merge operation on just the first element
				if ( DISPLAY) {
					cout << "merged set: " << endl;
					for ( unsigned int i = 0; i < merged.size(); ++i) {
						cout << merged[i] << ".";
					}
					cout << endl;
				}
				helper tree_guide(id, merged); //!<  add merged tree temp_clusters
				temp_clusters.push_back(tree_guide);
			}
			++sorter; //!<  go to next element in leaf
		}
		++rake; //!<  go to next leaf
	}

    if ( DISPLAY ) { //!<  display contents of the now merged taxa (optional)
		for ( unsigned int i = 0; i < temp_clusters.size(); ++i ) {
			temp_clusters[i].display();
		}
	}

	create_subtrees(temp_clusters, subtrees); //!<  make these merged taxa into actual subtrees
    cout << "sub-tree creation complete." << endl;

	//!<  merge subtrees together to make supertree: use hash_table
	map<vector<short>, list<SuperTree> > merge_hash;
	int longest = 0;
	merge_hash = create_superhash(subtrees, longest); //!<  create hash table
	//!<  display contents of hash table (optional)
	if ( DISPLAY ) {
		cout << "longest id is: " << longest << endl;
		cout << "Printing out merge_hash!" << endl;
		display_hash(merge_hash);
	}
	merge_supertrees(merge_hash, longest); //!<  merge subtrees together into final supertree

    return;
}

#endif
